@@ -32,10 +32,19 @@ class Parser(object):
         self.lookups_ = SymbolTable()
         self.next_token_type_, self.next_token_ = (None, None)
         self.next_token_location_ = None
-        with open(path, "r") as f:
-            self.lexer_ = Lexer(f.read(), path)
+        self.make_lexer_(path)
         self.advance_lexer_()
 
+    def make_lexer_(self, file_or_path):
+        if hasattr(file_or_path, "read"):
+            filename = getattr(file_or_path, "name", None)
+            data = file_or_path.read()
+        else:
+            filename = file_or_path
+            with open(file_or_path, "r") as f:
+                data = f.read()
+        self.lexer_ = Lexer(data, filename)
+
     def parse(self):
         statements = self.doc_.statements
         while self.next_token_type_ is not None:
@@ -44,10 +53,7 @@ class Parser(object):
                 func = getattr(self, PARSE_FUNCS[self.cur_token_])
                 statements.append(func())
             elif self.is_cur_keyword_("END"):
-                if self.next_token_type_ is not None:
-                    raise VoltLibError("Expected the end of the file",
-                                       self.cur_token_location_)
-                return self.doc_
+                break
             else:
                 raise VoltLibError(
                     "Expected " + ", ".join(sorted(PARSE_FUNCS.keys())),
@@ -76,7 +82,7 @@ class Parser(object):
         if self.next_token_ == "TYPE":
             self.expect_keyword_("TYPE")
             gtype = self.expect_name_()
-            assert gtype in ("BASE", "LIGATURE", "MARK")
+            assert gtype in ("BASE", "LIGATURE", "MARK", "COMPONENT")
         components = None
         if self.next_token_ == "COMPONENTS":
             self.expect_keyword_("COMPONENTS")
@@ -87,8 +93,9 @@ class Parser(object):
                 'Glyph "%s" (gid %i) already defined' % (name, gid),
                 location
             )
-        def_glyph = ast.GlyphDefinition(location, name, gid,
-                                        gunicode, gtype, components)
+        def_glyph = ast.GlyphDefinition(name, gid,
+                                        gunicode, gtype, components,
+                                        location=location)
         self.glyphs_.define(name, def_glyph)
         return def_glyph
 
@@ -98,7 +105,6 @@ class Parser(object):
         name = self.expect_string_()
         enum = None
         if self.next_token_ == "ENUM":
-            self.expect_keyword_("ENUM")
             enum = self.parse_enum_()
         self.expect_keyword_("END_GROUP")
         if self.groups_.resolve(name) is not None:
@@ -107,7 +113,8 @@ class Parser(object):
                 'group names are case insensitive' % name,
                 location
             )
-        def_group = ast.GroupDefinition(location, name, enum)
+        def_group = ast.GroupDefinition(name, enum,
+                                        location=location)
         self.groups_.define(name, def_group)
         return def_group
 
@@ -142,7 +149,7 @@ class Parser(object):
             langs.append(lang)
         self.expect_keyword_("END_SCRIPT")
         self.langs_.exit_scope()
-        def_script = ast.ScriptDefinition(location, name, tag, langs)
+        def_script = ast.ScriptDefinition(name, tag, langs, location=location)
         self.scripts_.define(tag, def_script)
         return def_script
 
@@ -161,7 +168,8 @@ class Parser(object):
             feature = self.parse_feature_()
             self.expect_keyword_("END_FEATURE")
             features.append(feature)
-        def_langsys = ast.LangSysDefinition(location, name, tag, features)
+        def_langsys = ast.LangSysDefinition(name, tag, features,
+                                            location=location)
         return def_langsys
 
     def parse_feature_(self):
@@ -177,7 +185,8 @@ class Parser(object):
             self.expect_keyword_("LOOKUP")
             lookup = self.expect_string_()
             lookups.append(lookup)
-        feature = ast.FeatureDefinition(location, name, tag, lookups)
+        feature = ast.FeatureDefinition(name, tag, lookups,
+                                        location=location)
         return feature
 
     def parse_def_lookup_(self):
@@ -202,11 +211,12 @@ class Parser(object):
             self.advance_lexer_()
             process_base = False
         process_marks = True
+        mark_glyph_set = None
         if self.next_token_ == "PROCESS_MARKS":
             self.advance_lexer_()
             if self.next_token_ == "MARK_GLYPH_SET":
                 self.advance_lexer_()
-                process_marks = self.expect_string_()
+                mark_glyph_set = self.expect_string_()
             elif self.next_token_type_ == Lexer.STRING:
                 process_marks = self.expect_string_()
             elif self.next_token_ == "ALL":
@@ -248,8 +258,8 @@ class Parser(object):
                 "Got %s" % (as_pos_or_sub),
                 location)
         def_lookup = ast.LookupDefinition(
-            location, name, process_base, process_marks, direction, reversal,
-            comments, context, sub, pos)
+            name, process_base, process_marks, mark_glyph_set, direction,
+            reversal, comments, context, sub, pos, location=location)
         self.lookups_.define(name, def_lookup)
         return def_lookup
 
@@ -272,8 +282,8 @@ class Parser(object):
                     else:
                         right.append(coverage)
                 self.expect_keyword_("END_CONTEXT")
-                context = ast.ContextDefinition(location, ex_or_in, left,
-                                                right)
+                context = ast.ContextDefinition(ex_or_in, left,
+                                                right, location=location)
                 contexts.append(context)
             else:
                 self.expect_keyword_("END_CONTEXT")
@@ -305,13 +315,16 @@ class Parser(object):
         if max_src == 1 and max_dest == 1:
             if reversal:
                 sub = ast.SubstitutionReverseChainingSingleDefinition(
-                    location, mapping)
+                    mapping, location=location)
             else:
-                sub = ast.SubstitutionSingleDefinition(location, mapping)
+                sub = ast.SubstitutionSingleDefinition(mapping,
+                                                       location=location)
         elif max_src == 1 and max_dest > 1:
-            sub = ast.SubstitutionMultipleDefinition(location, mapping)
+            sub = ast.SubstitutionMultipleDefinition(mapping,
+                                                     location=location)
         elif max_src > 1 and max_dest == 1:
-            sub = ast.SubstitutionLigatureDefinition(location, mapping)
+            sub = ast.SubstitutionLigatureDefinition(mapping,
+                                                     location=location)
         return sub
 
     def parse_position_(self):
@@ -348,7 +361,7 @@ class Parser(object):
             coverage_to.append((cov, anchor_name))
         self.expect_keyword_("END_ATTACH")
         position = ast.PositionAttachDefinition(
-            location, coverage, coverage_to)
+            coverage, coverage_to, location=location)
         return position
 
     def parse_attach_cursive_(self):
@@ -364,7 +377,7 @@ class Parser(object):
             coverages_enter.append(self.parse_coverage_())
         self.expect_keyword_("END_ATTACH")
         position = ast.PositionAttachCursiveDefinition(
-            location, coverages_exit, coverages_enter)
+            coverages_exit, coverages_enter, location=location)
         return position
 
     def parse_adjust_pair_(self):
@@ -390,7 +403,7 @@ class Parser(object):
             adjust_pair[(id_1, id_2)] = (pos_1, pos_2)
         self.expect_keyword_("END_ADJUST")
         position = ast.PositionAdjustPairDefinition(
-            location, coverages_1, coverages_2, adjust_pair)
+            coverages_1, coverages_2, adjust_pair, location=location)
         return position
 
     def parse_adjust_single_(self):
@@ -404,7 +417,7 @@ class Parser(object):
             adjust_single.append((coverages, pos))
         self.expect_keyword_("END_ADJUST")
         position = ast.PositionAdjustSingleDefinition(
-            location, adjust_single)
+            adjust_single, location=location)
         return position
 
     def parse_def_anchor_(self):
@@ -415,16 +428,17 @@ class Parser(object):
         gid = self.expect_number_()
         self.expect_keyword_("GLYPH")
         glyph_name = self.expect_name_()
-        # check for duplicate anchor names on this glyph
-        if (glyph_name in self.anchors_
-                and self.anchors_[glyph_name].resolve(name) is not None):
-            raise VoltLibError(
-                'Anchor "%s" already defined, '
-                'anchor names are case insensitive' % name,
-                location
-            )
         self.expect_keyword_("COMPONENT")
         component = self.expect_number_()
+        # check for duplicate anchor names on this glyph
+        if glyph_name in self.anchors_:
+            anchor = self.anchors_[glyph_name].resolve(name)
+            if anchor is not None and anchor.component == component:
+                raise VoltLibError(
+                    'Anchor "%s" already defined, '
+                    'anchor names are case insensitive' % name,
+                    location
+                )
         if self.next_token_ == "LOCKED":
             locked = True
             self.advance_lexer_()
@@ -433,8 +447,9 @@ class Parser(object):
         self.expect_keyword_("AT")
         pos = self.parse_pos_()
         self.expect_keyword_("END_ANCHOR")
-        anchor = ast.AnchorDefinition(location, name, gid, glyph_name,
-                                      component, locked, pos)
+        anchor = ast.AnchorDefinition(name, gid, glyph_name,
+                                      component, locked, pos,
+                                      location=location)
         if glyph_name not in self.anchors_:
             self.anchors_[glyph_name] = SymbolTable()
         self.anchors_[glyph_name].define(name, anchor)
@@ -492,9 +507,9 @@ class Parser(object):
         return unicode_values if unicode_values != [] else None
 
     def parse_enum_(self):
-        assert self.is_cur_keyword_("ENUM")
+        self.expect_keyword_("ENUM")
         location = self.cur_token_location_
-        enum = self.parse_coverage_()
+        enum = ast.Enum(self.parse_coverage_(), location=location)
         self.expect_keyword_("END_ENUM")
         return enum
 
@@ -503,55 +518,42 @@ class Parser(object):
         location = self.cur_token_location_
         while self.next_token_ in ("GLYPH", "GROUP", "RANGE", "ENUM"):
             if self.next_token_ == "ENUM":
-                self.advance_lexer_()
                 enum = self.parse_enum_()
                 coverage.append(enum)
             elif self.next_token_ == "GLYPH":
                 self.expect_keyword_("GLYPH")
                 name = self.expect_string_()
-                coverage.append(name)
+                coverage.append(ast.GlyphName(name, location=location))
             elif self.next_token_ == "GROUP":
                 self.expect_keyword_("GROUP")
                 name = self.expect_string_()
-                # resolved_group = self.groups_.resolve(name)
-                group = (name,)
-                coverage.append(group)
-                # if resolved_group is not None:
-                #     coverage.extend(resolved_group.enum)
-                # # TODO: check that group exists after all groups are defined
-                # else:
-                #     group = (name,)
-                #     coverage.append(group)
-                #     # raise VoltLibError(
-                #     #     'Glyph group "%s" is not defined' % name,
-                #     #     location)
+                coverage.append(ast.GroupName(name, self, location=location))
             elif self.next_token_ == "RANGE":
                 self.expect_keyword_("RANGE")
                 start = self.expect_string_()
                 self.expect_keyword_("TO")
                 end = self.expect_string_()
-                coverage.append((start, end))
+                coverage.append(ast.Range(start, end, self, location=location))
         return tuple(coverage)
 
     def resolve_group(self, group_name):
         return self.groups_.resolve(group_name)
 
     def glyph_range(self, start, end):
-        rng = self.glyphs_.range(start, end)
-        return frozenset(rng)
+        return self.glyphs_.range(start, end)
 
     def parse_ppem_(self):
         location = self.cur_token_location_
         ppem_name = self.cur_token_
         value = self.expect_number_()
-        setting = ast.SettingDefinition(location, ppem_name, value)
+        setting = ast.SettingDefinition(ppem_name, value, location=location)
         return setting
 
     def parse_compiler_flag_(self):
         location = self.cur_token_location_
         flag_name = self.cur_token_
         value = True
-        setting = ast.SettingDefinition(location, flag_name, value)
+        setting = ast.SettingDefinition(flag_name, value, location=location)
         return setting
 
     def parse_cmap_format(self):
@@ -559,7 +561,7 @@ class Parser(object):
         name = self.cur_token_
         value = (self.expect_number_(), self.expect_number_(),
                  self.expect_number_())
-        setting = ast.SettingDefinition(location, name, value)
+        setting = ast.SettingDefinition(name, value, location=location)
         return setting
 
     def is_cur_keyword_(self, k):
@@ -594,6 +596,8 @@ class Parser(object):
         self.cur_token_type_, self.cur_token_, self.cur_token_location_ = (
             self.next_token_type_, self.next_token_, self.next_token_location_)
         try:
+            if self.is_cur_keyword_("END"):
+                raise StopIteration
             (self.next_token_type_, self.next_token_,
              self.next_token_location_) = self.lexer_.next()
         except StopIteration:
